
from .api import AIONetbox

__title__ = 'aionetbox'
__version__ = '1.3.0'
__description__ = 'Async netbox library'
__author__ = 'Marco Ceppi'
__author_email__ = 'marco@vapor.io'
__url__ = 'https://github.com/vapor-ware/aionetbox'
__license__ = 'GNU General Public License v3.0'
